@@ -2,18 +2,14 @@
 Mount utilities.
 """
 
-from __future__ import absolute_import
-from __future__ import unicode_literals
-
 from distutils.spawn import find_executable
 from collections import namedtuple
 from functools import partial
-import inspect
 import logging
 import os
 
-from .async_ import AsyncList, Coroutine, Return, sleep
-from .common import wraps, setdefault, exc_message
+from .async_ import to_coro, gather, sleep
+from .common import wraps, setdefault, exc_message, format_exc
 from .config import IgnoreDevice, match_config
 from .locale import _
 
@@ -24,70 +20,16 @@ __all__ = ['Mounter']
 # TODO: add / remove / XXX_all should make proper use of the asynchronous
 # execution.
 
-
-@Coroutine.from_generator_function
-def _False():
-    yield Return(False)
-
-
-def _find_device(fn):
-    """
-    Decorator for Mounter methods taking a Device as their first argument.
-
-    Enables to pass the path name as first argument and does some common error
-    handling (logging).
-    """
+def _error_boundary(fn):
     @wraps(fn)
-    def wrapper(self, device_or_path, *args, **kwargs):
+    async def wrapper(self, device, *args, **kwargs):
         try:
-            device = self.udisks.find(device_or_path)
-        except ValueError as e:
-            self._log.error(exc_message(e))
-            return _False()
-        return Coroutine(fn(self, device, *args, **kwargs))
-    return wrapper
-
-
-def _find_device_auto_losetup(fn):
-    @Coroutine.from_generator_function
-    def wrapper(self, device_or_path, *args, **kwargs):
-        try:
-            try:
-                device = self.udisks.find(device_or_path)
-            except ValueError as e:
-                if not os.path.isfile(device_or_path):
-                    raise
-                device = yield self.losetup(device_or_path)
-                bound = inspect.getcallargs(fn, self, device, *args, **kwargs)
-                if bound.get('recursive') is False:
-                    yield Return(device)
+            return await fn(self, device, *args, **kwargs)
         except Exception as e:
-            self._log.error(exc_message(e))
-            yield Return(False)
-        result = yield Coroutine(fn(self, device, *args, **kwargs))
-        yield Return(result)
-    return wrapper
-
-
-def _sets_async_error(fn):
-    @wraps(fn)
-    def wrapper(self, device, *args, **kwargs):
-        async_ = fn(self, device, *args, **kwargs)
-        async_.errbacks.append(partial(self._error, fn, device))
-        return async_
-    return wrapper
-
-
-def _suppress_error(fn):
-    """
-    Prevent errors in this function from being shown. This is OK, since all
-    errors happen in sub-functions in which errors ARE logged.
-    """
-    @wraps(fn)
-    def wrapper(self, device, *args, **kwargs):
-        async_ = fn(self, device, *args, **kwargs)
-        async_.errbacks.append(lambda *args: True)
-        return async_
+            self._log.error(_('failed to {0} {1}: {2}',
+                              fn.__name__, device, exc_message(e)))
+            self._log.debug(format_exc())
+            return False
     return wrapper
 
 
@@ -100,12 +42,7 @@ def _is_parent_of(parent, child):
     return False
 
 
-def _get_parent(device):
-    """Return the container device or ``None``."""
-    return device.partition_slave or device.luks_cleartext_slave
-
-
-class Mounter(object):
+class Mounter:
 
     """
     Mount utility.
@@ -119,11 +56,8 @@ class Mounter(object):
     should always be passed as keyword arguments.
     """
 
-    def __init__(self, udisks,
-                 config=None,
-                 prompt=None,
-                 browser=None,
-                 cache=None):
+    def __init__(self, udisks, config=None, prompt=None, browser=None,
+                 terminal=None, cache=None, cache_hint=False):
         """
         Initialize mounter with the given defaults.
 
@@ -131,6 +65,7 @@ class Mounter(object):
         :param list config: list of :class:`DeviceFilter`
         :param callable prompt: retrieve passwords for devices
         :param callable browser: open devices
+        :param callable terminal: open devices in terminal
 
         If prompt is None, device unlocking will not work.
         If browser is None, browse will not work.
@@ -138,74 +73,98 @@ class Mounter(object):
         self.udisks = udisks
         self._config = (config or []) + [
             IgnoreDevice({'symlinks': '/dev/mapper/docker-*', 'ignore': True}),
-            IgnoreDevice({'symlinks': '/dev/disk/by-id/dm-name-docker-*', 'ignore': True}),
-            IgnoreDevice({'is_loop': True, 'loop_file': '/*', 'ignore': False}),
+            IgnoreDevice({'symlinks': '/dev/disk/by-id/dm-name-docker-*',
+                          'ignore': True}),
+            IgnoreDevice({'is_loop': True, 'is_ignored': False,
+                          'loop_file': '/*', 'ignore': False}),
             IgnoreDevice({'is_block': False, 'ignore': True}),
-            IgnoreDevice({'is_external': False, 'ignore': True}),
+            IgnoreDevice({'is_external': False,
+                          'is_toplevel': True, 'ignore': True}),
             IgnoreDevice({'is_ignored': True, 'ignore': True})]
         self._prompt = prompt
         self._browser = browser
+        self._terminal = terminal
         self._cache = cache
+        self._cache_hint = cache_hint
         self._log = logging.getLogger(__name__)
+
+    def _find_device(self, device_or_path):
+        """Find device object from path."""
+        return self.udisks.find(device_or_path)
+
+    async def _find_device_losetup(self, device_or_path):
         try:
-            # propagate error messages to UDisks1 daemon for 'Job failed'
-            # notifications.
-            self._set_error = self.udisks.set_error
-        except AttributeError:
-            self._set_error = lambda device, action, message: None
+            device = self.udisks.find(device_or_path)
+            return device, False
+        except FileNotFoundError:
+            if not os.path.isfile(device_or_path):
+                raise
+        device = await self.losetup(device_or_path)
+        return device, True
 
-    def _error(self, fn, device, err, fmt):
-        message = exc_message(err)
-        self._log.error(_('failed to {0} {1}: {2}',
-                          fn.__name__, device, message))
-        self._set_error(device, fn.__name__, message)
-        return True
-
-    @_sets_async_error
-    @_find_device
-    def browse(self, device):
+    @_error_boundary
+    async def browse(self, device):
         """
-        Browse device.
+        Launch file manager on the mount path of the specified device.
 
         :param device: device object, block device path or mount path
-        :returns: success
-        :rtype: bool
+        :returns: whether the program was successfully launched.
         """
+        device = self._find_device(device)
         if not device.is_mounted:
             self._log.error(_("not browsing {0}: not mounted", device))
-            yield Return(False)
+            return False
         if not self._browser:
             self._log.error(_("not browsing {0}: no program", device))
-            yield Return(False)
+            return False
         self._log.debug(_('opening {0} on {0.mount_paths[0]}', device))
         self._browser(device.mount_paths[0])
         self._log.info(_('opened {0} on {0.mount_paths[0]}', device))
-        yield Return(True)
+        return True
+
+    @_error_boundary
+    async def terminal(self, device):
+        """
+        Launch terminal on the mount path of the specified device.
+
+        :param device: device object, block device path or mount path
+        :returns: whether the program was successfully launched.
+        """
+        device = self._find_device(device)
+        if not device.is_mounted:
+            self._log.error(_("not opening terminal {0}: not mounted", device))
+            return False
+        if not self._terminal:
+            self._log.error(_("not opening terminal {0}: no program", device))
+            return False
+        self._log.debug(_('opening {0} on {0.mount_paths[0]}', device))
+        self._terminal(device.mount_paths[0])
+        self._log.info(_('opened {0} on {0.mount_paths[0]}', device))
+        return True
 
     # mount/unmount
-    @_sets_async_error
-    @_find_device
-    def mount(self, device):
+    @_error_boundary
+    async def mount(self, device):
         """
         Mount the device if not already mounted.
 
         :param device: device object, block device path or mount path
         :returns: whether the device is mounted.
-        :rtype: bool
         """
+        device = self._find_device(device)
         if not self.is_handleable(device) or not device.is_filesystem:
             self._log.warn(_('not mounting {0}: unhandled device', device))
-            yield Return(False)
+            return False
         if device.is_mounted:
             self._log.info(_('not mounting {0}: already mounted', device))
-            yield Return(True)
+            return True
         options = match_config(self._config, device, 'options', None)
         kwargs = dict(options=options)
         self._log.debug(_('mounting {0} with {1}', device, kwargs))
         self._check_device_before_mount(device)
-        mount_path = yield device.mount(**kwargs)
+        mount_path = await device.mount(**kwargs)
         self._log.info(_('mounted {0} on {1}', device, mount_path))
-        yield Return(True)
+        return True
 
     def _check_device_before_mount(self, device):
         if device.id_type == 'ntfs' and not find_executable('ntfs-3g'):
@@ -214,111 +173,118 @@ class Mounter(object):
                 "Please install 'ntfs-3g' if you experience problems or the "
                 "device is readonly."))
 
-    @_sets_async_error
-    @_find_device
-    def unmount(self, device):
+    @_error_boundary
+    async def unmount(self, device):
         """
         Unmount a Device if mounted.
 
         :param device: device object, block device path or mount path
         :returns: whether the device is unmounted
-        :rtype: bool
         """
+        device = self._find_device(device)
         if not self.is_handleable(device) or not device.is_filesystem:
             self._log.warn(_('not unmounting {0}: unhandled device', device))
-            yield Return(False)
+            return False
         if not device.is_mounted:
             self._log.info(_('not unmounting {0}: not mounted', device))
-            yield Return(True)
+            return True
         self._log.debug(_('unmounting {0}', device))
-        yield device.unmount()
+        await device.unmount()
         self._log.info(_('unmounted {0}', device))
-        yield Return(True)
+        return True
 
     # unlock/lock (LUKS)
-    @_sets_async_error
-    @_find_device
-    def unlock(self, device):
+    @_error_boundary
+    async def unlock(self, device):
         """
         Unlock the device if not already unlocked.
 
         :param device: device object, block device path or mount path
         :returns: whether the device is unlocked
-        :rtype: bool
         """
+        device = self._find_device(device)
         if not self.is_handleable(device) or not device.is_crypto:
             self._log.warn(_('not unlocking {0}: unhandled device', device))
-            yield Return(False)
+            return False
         if device.is_unlocked:
             self._log.info(_('not unlocking {0}: already unlocked', device))
-            yield Return(True)
+            return True
         if not self._prompt:
             self._log.error(_('not unlocking {0}: no password prompt', device))
-            yield Return(False)
-        unlocked = yield self._unlock_from_cache(device)
+            return False
+        unlocked = await self._unlock_from_cache(device)
         if unlocked:
-            yield Return(True)
-        unlocked = yield self._unlock_from_keyfile(device)
+            return True
+        unlocked = await self._unlock_from_keyfile(device)
         if unlocked:
-            yield Return(True)
-        password = yield self._prompt(device, self.udisks.keyfile_support)
+            return True
+        options = dict(allow_keyfile=self.udisks.keyfile_support,
+                       allow_cache=self._cache is not None,
+                       cache_hint=self._cache_hint)
+        password = await self._prompt(device, options)
+        # password is either None or udiskie.prompt.PasswordResult:
         if password is None:
             self._log.debug(_('not unlocking {0}: cancelled by user', device))
-            yield Return(False)
+            return False
+        cache_hint = password.cache_hint
+        password = password.password
         if isinstance(password, bytes):
             self._log.debug(_('unlocking {0} using keyfile', device))
-            yield device.unlock_keyfile(password)
+            await device.unlock_keyfile(password)
         else:
             self._log.debug(_('unlocking {0}', device))
-            yield device.unlock(password)
-        self._update_cache(device, password)
+            await device.unlock(password)
+        self._update_cache(device, password, cache_hint)
         self._log.info(_('unlocked {0}', device))
-        yield Return(True)
+        return True
 
-    @Coroutine.from_generator_function
-    def _unlock_from_cache(self, device):
+    async def _unlock_from_cache(self, device):
         if not self._cache:
-            yield Return(False)
+            return False
         try:
             password = self._cache[device]
         except KeyError:
-            yield Return(False)
+            self._log.debug(_("no cached key for {0}", device))
+            return False
         self._log.debug(_('unlocking {0} using cached password', device))
         try:
-            yield device.unlock(password)
+            await device.unlock_keyfile(password)
         except Exception:
             self._log.debug(_('failed to unlock {0} using cached password', device))
-            yield Return(False)
+            self._log.debug(format_exc())
+            return False
         self._log.info(_('unlocked {0} using cached password', device))
-        yield Return(True)
+        return True
 
-    @Coroutine.from_generator_function
-    def _unlock_from_keyfile(self, device):
+    async def _unlock_from_keyfile(self, device):
         if not self.udisks.keyfile_support:
-            yield Return(False)
+            return False
         filename = match_config(self._config, device, 'keyfile', None)
         if filename is None:
             self._log.debug(_('No matching keyfile rule for {}.', device))
-            yield Return(False)
+            return False
         try:
             with open(filename, 'rb') as f:
                 keyfile = f.read()
         except IOError:
-            self._log.warn(_('configured keyfile for {0} not found', device))
-            yield Return(False)
+            self._log.warn(_('keyfile for {0} not found: {1}', device, filename))
+            return False
         self._log.debug(_('unlocking {0} using keyfile {1}', device, filename))
         try:
-            yield device.unlock_keyfile(keyfile)
+            await device.unlock_keyfile(keyfile)
         except Exception:
             self._log.debug(_('failed to unlock {0} using keyfile', device))
-            yield Return(False)
+            self._log.debug(format_exc())
+            return False
         self._log.info(_('unlocked {0} using keyfile', device))
-        yield Return(True)
+        return True
 
-    def _update_cache(self, device, password):
+    def _update_cache(self, device, password, cache_hint):
         if not self._cache:
             return
-        self._cache[device] = password
+        # TODO: could allow numeric cache_hint (=timeout)…
+        if cache_hint or cache_hint is None:
+            self._cache[device] = password
 
     def forget_password(self, device):
         try:
@@ -326,47 +292,47 @@ class Mounter(object):
         except KeyError:
             pass
 
-    @_sets_async_error
-    @_find_device
-    def lock(self, device):
+    @_error_boundary
+    async def lock(self, device):
         """
         Lock device if unlocked.
 
         :param device: device object, block device path or mount path
         :returns: whether the device is locked
-        :rtype: bool
         """
+        device = self._find_device(device)
         if not self.is_handleable(device) or not device.is_crypto:
             self._log.warn(_('not locking {0}: unhandled device', device))
-            yield Return(False)
+            return False
         if not device.is_unlocked:
             self._log.info(_('not locking {0}: not unlocked', device))
-            yield Return(True)
+            return True
         self._log.debug(_('locking {0}', device))
-        yield device.lock()
+        await device.lock()
         self._log.info(_('locked {0}', device))
-        yield Return(True)
+        return True
 
     # add/remove (unlock/lock or mount/unmount)
-    @_suppress_error
-    @_find_device_auto_losetup
-    def add(self, device, recursive=None):
+    @_error_boundary
+    async def add(self, device, recursive=None):
         """
         Mount or unlock the device depending on its type.
 
         :param device: device object, block device path or mount path
         :param bool recursive: recursively mount and unlock child devices
         :returns: whether all attempted operations succeeded
-        :rtype: bool
         """
+        device, created = await self._find_device_losetup(device)
+        if created and recursive is False:
+            return device
         if device.is_filesystem:
-            success = yield self.mount(device)
+            success = await self.mount(device)
         elif device.is_crypto:
-            success = yield self.unlock(device)
+            success = await self.unlock(device)
             if success and recursive:
-                yield self.udisks._sync()
+                await self.udisks._sync()
                 device = self.udisks[device.object_path]
-                success = yield self.add(
+                success = await self.add(
                     device.luks_cleartext_holder,
                     recursive=True)
         elif (recursive
@@ -377,16 +343,15 @@ class Mounter(object):
                 for dev in self.get_all_handleable()
                 if dev.is_partition and dev.partition_slave == device
             ]
-            results = yield AsyncList(tasks)
+            results = await gather(*tasks)
             success = all(results)
         else:
             self._log.info(_('not adding {0}: unhandled device', device))
-            yield Return(False)
-        yield Return(success)
+            return False
+        return success
 
-    @_suppress_error
-    @_find_device_auto_losetup
-    def auto_add(self, device, recursive=None):
+    @_error_boundary
+    async def auto_add(self, device, recursive=None, automount=True):
         """
         Automatically attempt to mount or unlock a device, but be quiet if the
         device is not supported.
@@ -394,23 +359,25 @@ class Mounter(object):
         :param device: device object, block device path or mount path
         :param bool recursive: recursively mount and unlock child devices
         :returns: whether all attempted operations succeeded
-        :rtype: bool
         """
+        device, created = await self._find_device_losetup(device)
+        if created and recursive is False:
+            return device
+        if device.is_luks_cleartext and self.udisks.version_info >= (2, 7, 0):
+            await sleep(1.5)    # temporary workaround for #153, unreliable
         success = True
-        if device.is_luks_cleartext and self.udisks.version_info >= (2,7,0):
-            yield sleep(1.5)    # temporary workaround for #153, unreliable...
-        if not self.is_automount(device):
+        if not self.is_automount(device, automount):
             pass
         elif device.is_filesystem:
             if not device.is_mounted:
-                success = yield self.mount(device)
+                success = await self.mount(device)
         elif device.is_crypto:
             if self._prompt and not device.is_unlocked:
-                success = yield self.unlock(device)
+                success = await self.unlock(device)
             if success and recursive:
-                yield self.udisks._sync()
+                await self.udisks._sync()
                 device = self.udisks[device.object_path]
-                success = yield self.auto_add(
+                success = await self.auto_add(
                     device.luks_cleartext_holder,
                     recursive=True)
         elif recursive and device.is_partition_table:
@@ -419,16 +386,15 @@ class Mounter(object):
                 for dev in self.get_all_handleable()
                 if dev.is_partition and dev.partition_slave == device
             ]
-            results = yield AsyncList(tasks)
+            results = await gather(*tasks)
             success = all(results)
         else:
             self._log.debug(_('not adding {0}: unhandled device', device))
-        yield Return(success)
+        return success
 
-    @_suppress_error
-    @_find_device
-    def remove(self, device, force=False, detach=False, eject=False,
-               lock=False):
+    @_error_boundary
+    async def remove(self, device, force=False, detach=False, eject=False,
+                     lock=False):
         """
         Unmount or lock the device depending on device type.
 
@@ -438,15 +404,15 @@ class Mounter(object):
         :param bool eject: remove media from the root drive
         :param bool lock: lock the associated LUKS cleartext slave
         :returns: whether all attempted operations succeeded
-        :rtype: bool
         """
+        device = self._find_device(device)
         if device.is_filesystem:
             if device.is_mounted or not device.is_loop or detach is False:
-                success = yield self.unmount(device)
+                success = await self.unmount(device)
         elif device.is_crypto:
             if force and device.is_unlocked:
-                yield self.auto_remove(device.luks_cleartext_holder, force=True)
-            success = yield self.lock(device)
+                await self.auto_remove(device.luks_cleartext_holder, force=True)
+            success = await self.lock(device)
         elif (force
               and (device.is_partition_table or device.is_drive)
               and self.is_handleable(device)):
@@ -456,7 +422,7 @@ class Mounter(object):
                 for child in self.get_all_handleable()
                 if _is_parent_of(device, child)
             ]
-            results = yield AsyncList(tasks)
+            results = await gather(*tasks)
             success = all(results)
         else:
             self._log.info(_('not removing {0}: unhandled device', device))
@@ -464,19 +430,19 @@ class Mounter(object):
         # if these operations work, everything is fine, we can return True:
         if lock and device.is_luks_cleartext:
             device = device.luks_cleartext_slave
-            success = yield self.lock(device)
+            if self.is_handleable(device):
+                success = await self.lock(device)
         if eject:
-            success = yield self.eject(device)
+            success = await self.eject(device)
         if (detach or detach is None) and device.is_loop:
-            success = yield self.delete(device, remove=False)
+            success = await self.delete(device, remove=False)
         elif detach:
-            success = yield self.detach(device)
-        yield Return(success)
+            success = await self.detach(device)
+        return success
 
-    @_suppress_error
-    @_find_device
-    def auto_remove(self, device, force=False, detach=False, eject=False,
-                    lock=False):
+    @_error_boundary
+    async def auto_remove(self, device, force=False, detach=False, eject=False,
+                          lock=False):
         """
         Unmount or lock the device depending on device type.
 
@@ -486,19 +452,19 @@ class Mounter(object):
         :param bool eject: remove media from the root drive
         :param bool lock: lock the associated LUKS cleartext slave
         :returns: whether all attempted operations succeeded
-        :rtype: bool
         """
+        device = self._find_device(device)
         success = True
         if not self.is_handleable(device):
             pass
         elif device.is_filesystem:
             if device.is_mounted:
-                success = yield self.unmount(device)
+                success = await self.unmount(device)
         elif device.is_crypto:
             if force and device.is_unlocked:
-                yield self.auto_remove(device.luks_cleartext_holder, force=True)
+                await self.auto_remove(device.luks_cleartext_holder, force=True)
             if device.is_unlocked:
-                success = yield self.lock(device)
+                success = await self.lock(device)
         elif force and (device.is_partition_table or device.is_drive):
             kw = dict(force=True, detach=detach, eject=eject, lock=lock)
             tasks = [
@@ -506,93 +472,88 @@ class Mounter(object):
                 for child in self.get_all_handleable()
                 if _is_parent_of(device, child)
             ]
-            results = yield AsyncList(tasks)
+            results = await gather(*tasks)
             success = all(results)
         else:
             self._log.debug(_('not removing {0}: unhandled device', device))
         # if these operations work, everything is fine, we can return True:
         if lock and device.is_luks_cleartext:
             device = device.luks_cleartext_slave
-            success = yield self.lock(device)
+            success = await self.lock(device)
         if eject and device.has_media:
-            success = yield self.eject(device)
+            success = await self.eject(device)
         if (detach or detach is None) and device.is_loop:
-            success = yield self.delete(device, remove=False)
+            success = await self.delete(device, remove=False)
         elif detach and device.is_detachable:
-            success = yield self.detach(device)
-        yield Return(success)
+            success = await self.detach(device)
+        return success
 
     # eject/detach device
-    @_sets_async_error
-    @_find_device
-    def eject(self, device, force=False):
+    @_error_boundary
+    async def eject(self, device, force=False):
         """
         Eject a device after unmounting all its mounted filesystems.
 
         :param device: device object, block device path or mount path
         :param bool force: remove child devices before trying to eject
         :returns: whether the operation succeeded
-        :rtype: bool
         """
+        device = self._find_device(device)
         if not self.is_handleable(device):
             self._log.warn(_('not ejecting {0}: unhandled device'))
-            yield Return(False)
+            return False
         drive = device.drive
         if not (drive.is_drive and drive.is_ejectable):
             self._log.warn(_('not ejecting {0}: drive not ejectable', drive))
-            yield Return(False)
+            return False
         if force:
             # Can't autoremove 'device.drive', because that will be filtered
             # due to block=False:
-            yield self.auto_remove(device.root, force=True)
+            await self.auto_remove(device.root, force=True)
         self._log.debug(_('ejecting {0}', device))
-        yield drive.eject()
+        await drive.eject()
         self._log.info(_('ejected {0}', device))
-        yield Return(True)
+        return True
 
-    @_sets_async_error
-    @_find_device
-    def detach(self, device, force=False):
+    @_error_boundary
+    async def detach(self, device, force=False):
         """
         Detach a device after unmounting all its mounted filesystems.
 
         :param device: device object, block device path or mount path
         :param bool force: remove child devices before trying to detach
         :returns: whether the operation succeeded
-        :rtype: bool
         """
+        device = self._find_device(device)
         if not self.is_handleable(device):
             self._log.warn(_('not detaching {0}: unhandled device', device))
-            yield Return(False)
+            return False
         drive = device.root
         if not drive.is_detachable:
             self._log.warn(_('not detaching {0}: drive not detachable', drive))
-            yield Return(False)
+            return False
         if force:
-            yield self.auto_remove(drive, force=True)
+            await self.auto_remove(drive, force=True)
         self._log.debug(_('detaching {0}', device))
-        yield drive.detach()
+        await drive.detach()
         self._log.info(_('detached {0}', device))
-        yield Return(True)
+        return True
 
     # mount_all/unmount_all
-    @Coroutine.from_generator_function
-    def add_all(self, recursive=False):
+    async def add_all(self, recursive=False):
         """
         Add all handleable devices that available at start.
 
         :param bool recursive: recursively mount and unlock child devices
         :returns: whether all attempted operations succeeded
-        :rtype: bool
         """
         tasks = [self.auto_add(device, recursive=recursive)
                  for device in self.get_all_handleable_leaves()]
-        results = yield AsyncList(tasks)
+        results = await gather(*tasks)
         success = all(results)
-        yield Return(success)
+        return success
 
-    @Coroutine.from_generator_function
-    def remove_all(self, detach=False, eject=False, lock=False):
+    async def remove_all(self, detach=False, eject=False, lock=False):
         """
         Remove all filesystems handleable by udiskie.
 
@@ -600,19 +561,17 @@ class Mounter(object):
         :param bool eject: remove media from the root drive
         :param bool lock: lock the associated LUKS cleartext slave
         :returns: whether all attempted operations succeeded
-        :rtype: bool
         """
         kw = dict(force=True, detach=detach, eject=eject, lock=lock)
         tasks = [self.auto_remove(device, **kw)
                  for device in self.get_all_handleable_roots()]
-        results = yield AsyncList(tasks)
+        results = await gather(*tasks)
         success = all(results)
-        yield Return(success)
+        return success
 
     # loop devices
-    @Coroutine.from_generator_function
-    def losetup(self, image,
-                read_only=True, offset=None, size=None, no_part_scan=None):
+    async def losetup(self, image, read_only=True, offset=None, size=None,
+                      no_part_scan=None):
         """
         Setup a loop device.
 
@@ -625,20 +584,17 @@ class Mounter(object):
         """
         try:
             device = self.udisks.find(image)
-        except ValueError:
+        except FileNotFoundError:
             pass
         else:
             self._log.info(_('not setting up {0}: already up', device))
-            yield Return(device)
+            return device
         if not os.path.isfile(image):
             self._log.error(_('not setting up {0}: not a file', image))
-            yield Return(None)
-        if not self.udisks.loop_support:
-            self._log.error(_('not setting up {0}: unsupported in UDisks1', image))
-            yield Return(None)
+            return None
         self._log.debug(_('setting up {0}', image))
         fd = os.open(image, os.O_RDONLY)
-        device = yield self.udisks.loop_setup(fd, {
+        device = await self.udisks.loop_setup(fd, {
             'offset': offset,
             'size': size,
             'read-only': read_only,
@@ -646,61 +602,53 @@ class Mounter(object):
         })
         self._log.info(_('set up {0} as {1}', image,
                          device.device_presentation))
-        yield Return(device)
+        return device
 
-    @_sets_async_error
-    @_find_device
-    def delete(self, device, remove=True):
+    @_error_boundary
+    async def delete(self, device, remove=True):
         """
         Detach the loop device.
 
         :param device: device object, block device path or mount path
         :param bool remove: whether to unmount the partition etc.
         :returns: whether the loop device is deleted
-        :rtype: bool
         """
+        device = self._find_device(device)
         if not self.is_handleable(device) or not device.is_loop:
             self._log.warn(_('not deleting {0}: unhandled device', device))
-            yield Return(False)
+            return False
         if remove:
-            yield self.auto_remove(device, force=True)
+            await self.auto_remove(device, force=True)
         self._log.debug(_('deleting {0}', device))
-        yield device.delete()
+        await device.delete()
         self._log.info(_('deleted {0}', device))
-        yield Return(True)
+        return True
 
     # iterate devices
     def is_handleable(self, device):
-        # TODO: handle pathes in first argument
+        # TODO: handle paths in first argument
         """
         Check whether this device should be handled by udiskie.
 
         :param device: device object, block device path or mount path
         :returns: handleability
-        :rtype: bool
 
         Currently this just means that the device is removable and holds a
         filesystem or the device is a LUKS encrypted volume.
         """
-        ignored = self._ignore_device(device)
-        # propagate handleability of parent devices:
-        if ignored is None and device is not None:
-            return self.is_handleable(_get_parent(device))
-        return not ignored
+        return not self._ignore_device(device)
 
-    def is_automount(self, device):
+    def is_automount(self, device, default=True):
         if not self.is_handleable(device):
             return False
-        return match_config(self._config, device, 'automount', True)
+        return match_config(self._config, device, 'automount', default)
 
     def _ignore_device(self, device):
         return match_config(self._config, device, 'ignore', False)
 
-    def is_addable(self, device):
-        """
-        Check if device can be added with ``auto_add``.
-        """
-        if not self.is_automount(device):
+    def is_addable(self, device, automount=True):
+        """Check if device can be added with ``auto_add``."""
+        if not self.is_automount(device, automount):
             return False
         if device.is_filesystem:
             return not device.is_mounted
@@ -713,9 +661,7 @@ class Mounter(object):
         return False
 
     def is_removable(self, device):
-        """
-        Check if device can be removed with ``auto_remove``.
-        """
+        """Check if device can be removed with ``auto_remove``."""
         if not self.is_handleable(device):
             return False
         if device.is_filesystem:
@@ -729,12 +675,7 @@ class Mounter(object):
         return False
 
     def get_all_handleable(self):
-        """
-        Enumerate all handleable devices currently known to udisks.
-
-        :returns: handleable devices
-        :rtype: list
-        """
+        """Get list of all known handleable devices."""
         nodes = self.get_device_tree()
         return [node.device
                 for node in sorted(nodes.values(), key=DevNode._sort_key)
@@ -775,6 +716,7 @@ class Mounter(object):
         device_nodes['/'] = root
         for node in device_nodes.values():
             node.children.sort(key=DevNode._sort_key)
+
         # use parent as fallback, update top->down:
         def propagate_ignored(node):
             for child in node.children:
@@ -784,23 +726,6 @@ class Mounter(object):
         propagate_ignored(root)
         return device_nodes
 
-    def get_device_tree_filtered(self):
-        """Get a tree of all handleable devices."""
-        root = self.get_device_tree()['/']
-        # remove ignored devices from the structure:
-        def filter_handled(node):
-            node.children[:] = [
-                descendant
-                for child in node.children
-                for descendant in filter_handled(child)
-            ]
-            if node.ignored:
-                return node.children
-            else:
-                return [dev]
-        filter_handled(root)
-        return root
-
 
 class DevNode:
 
@@ -819,10 +744,11 @@ Device = namedtuple('Device', ['root', 'branches', 'device', 'label', 'methods']
 Action = namedtuple('Action', ['method', 'device', 'label', 'action'])
 
 
-class DeviceActions(object):
+class DeviceActions:
 
     _labels = {
         'browse': _('Browse {0}'),
+        'terminal': _('Hack on {0}'),
         'mount': _('Mount {0}'),
         'unmount': _('Unmount {0}'),
         'unlock': _('Unlock {0}'),
@@ -838,13 +764,14 @@ class DeviceActions(object):
         self._actions = _actions = actions.copy()
         setdefault(_actions, {
             'browse': mounter.browse,
+            'terminal': mounter.terminal,
             'mount': mounter.mount,
             'unmount': mounter.unmount,
             'unlock': mounter.unlock,
             'lock': partial(mounter.remove, force=True),
             'eject': partial(mounter.eject, force=True),
             'detach': partial(mounter.detach, force=True),
-            'forget_password': mounter.forget_password,
+            'forget_password': to_coro(mounter.forget_password),
             'delete': mounter.delete,
         })
 
@@ -853,8 +780,7 @@ class DeviceActions(object):
         Detect all currently known devices.
 
         :param str root_device: object path of root device to return
-        :returns: root of device hierarchy
-        :rtype: Device
+        :returns: root node of device hierarchy
         """
         root = Device(None, [], None, "", [])
         device_nodes = dict(map(self._device_node,
@@ -873,6 +799,8 @@ class DeviceActions(object):
             if device.is_mounted:
                 if self._mounter._browser:
                     yield 'browse'
+                if self._mounter._terminal:
+                    yield 'terminal'
                 yield 'unmount'
             else:
                 yield 'mount'
@@ -888,7 +816,7 @@ class DeviceActions(object):
             yield 'eject'
         if device.is_detachable:
             yield 'detach'
-        if device.is_loop and device.loop_support:
+        if device.is_loop:
             yield 'delete'
 
     def _device_node(self, device):
